"""lg-rez / blocs / Outils divers et variés

Récupération d'objets Discord, décorateurs pour commandes, structures d'interaction dans les channels, utilitaires d'emojis, de date/temps, de formatage...

"""

from functools import wraps
import asyncio
import datetime
import unidecode
import re

import discord
import discord.utils
from discord.ext import commands

from lgrez.blocs import bdd, bdd_tools
from lgrez.blocs.bdd import Tables, Joueurs, Roles, BaseActions, Actions, BaseActionsRoles, Taches, Triggers, Reactions, CandidHaro
# on importe toutes les tables, plus simple pour y accéder depuis des réactions etc (via eval_accols)


### ---------------------------------------------------------------------------
### Utilitaires de récupération d'objets Discord (détectent les mentions)
### ---------------------------------------------------------------------------

#: Raccourci pour :func:`discord.utils.get`
get = discord.utils.get


def _find_by_mention_or_name(collec, val, pattern=None, must_be_found=False, raiser=None):
    """Utilitaire pour la suite : trouve <val> dans <collec>

    [pattern]           Motif RegEx à utiliser pour la recherche
    [must_be_found]     Si True, raise une AssertionError si <val> est introuvable
    [raiser]            Nom de la fonction à envoyer dans l'exception si introuvable
    """
    if not val:
        item = None
    elif pattern and (match := re.search(pattern, val)):
        item = get(collec, id=int(match.group(1)))
    else:
        item = get(collec, name=val)

    if must_be_found:
        assert item, f"{raiser or 'tools._find_by_mention_or_name'} : Élément {val} introuvable"

    return item


def channel(arg, nom, must_be_found=True):
    """Renvoie l'objet associé au salon ``#nom``.

    Args:
        nom (:class:`str`): nom du channel (texte/vocal/catégorie) ou sa mention (détection directe par regex)
        arg (:class:`~discord.ext.commands.Context` | :class:`~discord.Guild` | :class:`~discord.Member` | :class:`~discord.abc.GuildChannel`): argument "connecté" au serveur, permettant de remonter aux channels
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une :exc:`AssertionError` si le channel ``#nom`` n'existe pas (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.abc.GuildChannel`
    """
    try:
        channels = arg.channels if isinstance(arg, discord.Guild) else arg.guild.channels
    except AttributeError:
        raise TypeError("tools.channel : Impossible de remonter aux channels depuis l'argument trasmis")
    return _find_by_mention_or_name(channels, nom, pattern="<#([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.channel")


def role(arg, nom, must_be_found=True):
    """Renvoie l'objet associé au rôle ``@&nom``.

    Args:
        nom (:class:`str`): nom du rôle ou sa mention (détection directe par regex)
        arg (:class:`~discord.ext.commands.Context` | :class:`~discord.Guild` | :class:`~discord.Member` | :class:`~discord.abc.GuildChannel`): argument "connecté" au serveur, permettant de remonter aux rôles
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une :exc:`AssertionError` si le channel ``@&nom`` n'existe pas (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.Role`
    """
    try:
        roles = arg.roles if isinstance(arg, discord.Guild) else arg.guild.roles
    except AttributeError:
        raise TypeError("tools.role : Impossible de remonter aux rôles depuis l'argument trasmis")
    return _find_by_mention_or_name(roles, nom, pattern="<@&([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.role")


def member(arg, nom, must_be_found=True):
    """Renvoie l'objet associé au membre ``@nom``.

    Args:
        nom (:class:`str`): nom du joueur ou sa mention (détection directe par regex)
        arg (:class:`~discord.ext.commands.Context` | :class:`~discord.Guild` | :class:`~discord.Member` | :class:`~discord.abc.GuildChannel`): argument "connecté" au serveur, permettant de remonter aux membres
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une :exc:`AssertionError` si le membre ``@nom`` n'existe pas (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.Member`
    """
    try:
        members = arg.members if isinstance(arg, discord.Guild) else arg.guild.members
    except AttributeError:
        raise TypeError("tools.member : Impossible de remonter aux membres depuis l'argument trasmis")
    return _find_by_mention_or_name(members, nom, pattern="<@!([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.member")


def emoji(arg, nom, must_be_found=True):
    """Renvoie l'objet associé à l'emoji ``:nom:``.

    Args:
        nom (:class:`str`): nom de l'emoji (texte/vocal/catégorie) ou son utilisation (détection directe par regex)
        arg (:class:`~discord.ext.commands.Context` | :class:`~discord.Guild` | :class:`~discord.Member` | :class:`~discord.abc.GuildChannel`): argument "connecté" au serveur, permettant de remonter aux emojis
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une :exc:`AssertionError` si l'emoji ``:nom:`` n'existe pas (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.Emoji`
    """
    try:
        emojis = arg.emojis if isinstance(arg, discord.Guild) else arg.guild.emojis
    except AttributeError:
        raise TypeError("tools.emoji : Impossible de remonter aux emojis depuis l'argument trasmis")
    return _find_by_mention_or_name(emojis, nom, pattern="<:.*:([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.emoji")


def private_chan(member, must_be_found=True):
    """Renvie le channel privé d'un joueur

    Args:
        member (:class:`discord.Member`): membre du serveur
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une :exc:`AssertionError` si le channel n'existe pas (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.TextChannel`
    """
    joueur = Joueurs.query.get(member.id)
    assert joueur, f"tools.private_chan : Joueur {member} introuvable"
    chan = member.guild.get_channel(joueur.chan_id_)
    if must_be_found:
        assert chan, f"tools.private_chan : Chan privé de {joueur} introuvable"
    return chan


# Appel aux MJs
def mention_MJ(arg):
    """Renvoie la mention du rôle "MJ" si le joueur n'est pas un MJ, ``"@MJ"`` sinon.

    Args:
        arg (:class:`~discord.Member` | :class:`~discord.ext.commands.Context`): membre ou contexte d'un message envoyé par un membre
    """
    member = arg.author if hasattr(arg, "author") else arg
    if hasattr(member, "top_role") and member.top_role.name == "MJ":    # Si webhook, pas de top_role
        return "@MJ"
    else:
        return role(arg, "MJ").mention



### ---------------------------------------------------------------------------
### Exceptions
### ---------------------------------------------------------------------------

class CommandExit(RuntimeError):
    """Lever cette exception force l'arrêt immédiat d'une commande, et empêche le bot de réagir à nouveau

    Sous-classe de :exc:`RuntimeError`
    """
    pass


### ---------------------------------------------------------------------------
### Décorateurs pour les différentes commandes, en fonction de leur usage
### ---------------------------------------------------------------------------

#: décorateur pour commande (:func:`discord.ext.commands.check`) : commandes exécutables uniquement par un MJ ou un webhook
mjs_only = commands.check_any(commands.check(lambda ctx: ctx.message.webhook_id), commands.has_role("MJ"))

#: décorateur pour commande (:func:`discord.ext.commands.check`) : commandes exécutables par un MJ, un rédacteur ou un webhook (pour IA)
mjs_et_redacteurs = commands.check_any(commands.check(lambda ctx: ctx.message.webhook_id), commands.has_any_role("MJ", "Rédacteur"))

#: décorateur pour commande (:func:`discord.ext.commands.check`) : commandes exécutables uniquement par un joueur (inscrit en base), vivant ou mort
joueurs_only = commands.has_any_role("Joueur en vie", "Joueur mort")

#: décorateur pour commande (:func:`discord.ext.commands.check`)  : commandes exécutables uniquement par un joueur vivant
vivants_only = commands.has_role("Joueur en vie")

# @tools.private
def private(cmd):
    """Décorateur pour commande : lors d'une invocation de la commande décorée hors d'un channel privé (``#conv-bot-``), supprime le message d'invocation et exécute la commande dans le channel privée de l'invoqueur.

    Ce décorateur n'est utilisable que sur une commande définie dans un Cog.
    Si le joueur ayant utilisé la commande n'a pas de chan privé (pas en base), raise une :exc:`AssertionError`.

    Utilisable en combinaison avec :func:`.joueurs_only` et :func:`.vivants_only` (pas avec les autres attention, vu que seuls les joueurs ont un channel privé)
    """
    @wraps(cmd)
    async def new_cmd(self, ctx, *args, **kwargs):              # Cette commande est renvoyée à la place de cmd
        if not ctx.channel.name.startswith("conv-bot-"):            # Si pas déjà dans une conv bot :
            await ctx.message.delete()                                  # On supprime le message,
            ctx.channel = private_chan(ctx.author)                      # On remplace le chan dans le contexte d'appel par le chan privé,
            await ctx.send(f"{quote(ctx.message.content)}\n"            # On envoie un warning dans le chan privé,
                           f"{ctx.author.mention} :warning: Cette commande est interdite en dehors de ta conv privée ! :warning:\n"
                           f"J'ai supprimé ton message, et j'exécute la commande ici :")
        return await cmd(self, ctx, *args, **kwargs)                # Et on appelle cmd, avec le contexte modifié !

    return new_cmd



### ---------------------------------------------------------------------------
### Commandes d'interaction avec les joueurs : input, boucles, confirmation...
### ---------------------------------------------------------------------------

# Commande générale, à utiliser à la place de bot.wait_for('message', ...)
async def wait_for_message(bot, check, trigger_on_commands=False):
    """Attend et renvoie le premier message reçu rencontrant les critères demandés.

    Surcouche de :meth:`discord.ext.commands.Bot.wait_for` permettant d'ignorer les commandes et de réagir au mot-clé ``stop``.

    Args:
        bot (:class:`.LGBot`): bot connecté au serveur
        check (:class:`function`\(:class:`discord.Message` -> :class:`bool`\)): fonction validant ou non chaque message
        trigger_on_commands (:class:`bool`): si ``False`` (défaut), un message respectant ``check`` sera ignoré si c'est une commande.

    Returns:
        :class:`discord.Message`

    Si le message est ``"stop"`` ou ``"!stop"`` (ou autre casse), raise une exception :exc:`.CommandExit` (même si le message respecte ``check``).
    """
    if trigger_on_commands:
        def trig_check(m):
            return (check(m) or m.content.lower() in ["stop", "!stop"])         # et on trigger en cas de STOP
    else:
        def trig_check(m):
            return ((check(m)
                     and not m.content.startswith(bot.command_prefix))          # on ne trigger pas sur les commandes
                    or m.content.lower() in ["stop", "!stop"])                  # et on trigger en cas de STOP

    message = await bot.wait_for('message', check=trig_check)
    if message.content.lower() in ["stop", "!stop"]:
        if message.author == bot.user:
            raise CommandExit(ital("(Arrêt commande précédente)"))
        else:
            raise CommandExit("Arrêt demandé")
    else:
        return message


# Raccourci pratique
async def wait_for_message_here(ctx, trigger_on_commands=False):
    """Attend et renvoie le premier message reçu dans <ctx>.

    Surcouche de :func:`.wait_for_message` filtrant uniquement les messages envoyés dans ``ctx.channel`` par quelqu'un d'autre que le bot

    Args:
        ctx (:class:`discord.ext.commands.Context`): contexte d'une commande
        trigger_on_commands: passé directement à :func:`.wait_for_message`

    Returns:
        :class:`discord.Message`
    """
    def trig_check(message):
        return (message.channel == ctx.channel and message.author != ctx.bot.user)

    message = await wait_for_message(ctx.bot, check=trig_check, trigger_on_commands=trigger_on_commands)
    return message


# Permet de boucler question -> réponse tant que la réponse vérifie pas les critères nécessaires dans chan
async def boucle_message(bot, chan, in_message, condition_sortie, rep_message=None):
    """Permet de lancer une boucle question/réponse tant que la réponse ne vérifie pas une condition

    Args:
        bot (:class:`.LGBot`): bot connecté au serveur
        chan (:class:`discord.TextChannel`): channel dans lequel lancer la boucle
        condition_sortie (:class:`function`\(:class:`discord.Message` -> :class:`bool`\)): fonction validant ou non chaque message
        in_message (:class:`str`): si défini, message à envoyer avant la boucle
        rep_message (:class:`str`): si défini, permet de définir un message de boucle différent de ``in_message`` (identique si ``None``). Doit être défini si ``in_message`` n'est pas défini.

    Returns:
        :class:`discord.Message`
    """
    if not rep_message:
        rep_message = in_message
    if not rep_message:
        raise ValueError("tools.boucle_message : [in_message] ou [rep_message] doit être défini !")

    def check_chan(m): #C heck que le message soit envoyé par l'utilisateur et dans son channel perso
        return m.channel == chan and m.author != bot.user

    await chan.send(in_message)
    rep = await wait_for_message(bot, check_chan)
    while not condition_sortie(rep):
        await chan.send(rep_message)
        rep = await wait_for_message(bot, check_chan)

    return rep


async def boucle_query_joueur(ctx, cible=None, message=None, sensi=0.5):
    """Retourne un joueur (entrée de BDD) d'après son nom

    Args:
        ctx (:class:`discord.ext.commands.Context`): contexte d'une commande
        cible (:class:`str`): premier essai de cible (donnée par le joueur dans l'appel à une commande, par exemple)
        message (:class:`str`): si défini (et ``cible`` non définie), message à envoyer avant la boucle
        sensi (:class:`float`): sensibilité de la recherche (c.f. :func:`.bdd_tools.find_nearest`)

    Returns:
        :class:`.bdd.Joueurs`

    Attend que le joueur entre un nom de joueur, et boucle 5 fois au max (avant de l'insulter et de raise une erreur) pour chercher le plus proche joueur dans la table :class:`.bdd.Joueurs`.
    """
    if message and not cible:
        await ctx.send(message)

    for i in range(5):
        if i == 0 and cible:            # Au premier tour, si on a donné une cible
            rep = cible
        else:
            mess = await wait_for_message_here(ctx)
            rep = mess.content.strip("()[]{}<>")

        if id := ''.join([c for c in rep if c.isdigit()]):      # Si la chaîne contient un nombre, on l'extrait
            if joueur := Joueurs.query.get(int(id)):                # Si cet ID correspond à un utilisateur, on le récupère
                return joueur                                       # On a trouvé l'utilisateur !

        nearest = bdd_tools.find_nearest(rep, Joueurs, carac="nom", sensi=sensi, solo_si_parfait=False, match_first_word=True)     # Sinon, recherche au plus proche

        if not nearest:
            await ctx.send("Aucune entrée trouvée, merci de réessayer :")

        elif nearest[0][1] == 1 and not (len(nearest) > 1 and nearest[1][1] == 1):        # Si le score le plus haut est égal à 1...
            return nearest[0][0]        # ...renvoyer l'entrée correspondante

        elif len(nearest) == 1:
            m = await ctx.send(f"Je n'ai trouvé qu'une correspondance : {nearest[0][0].nom}\nÇa part ?")
            if await yes_no(ctx.bot, m):
                return nearest[0][0]
            else:
                await ctx.send("Bon d'accord, alors qui ?")

        else:
            s = "Les joueurs les plus proches de ton entrée sont les suivants : \n"
            for i, j in enumerate(nearest[:10]):
                s += f"{emoji_chiffre(i+1)}. {j[0].nom} \n"
            m = await ctx.send(s + ital("Tu peux les choisir en réagissant à ce message, ou en répondant au clavier."))
            n = await choice(ctx.bot, m, min(10, len(nearest)))
            return nearest[n-1][0]

    await ctx.send("Et puis non, tiens ! \n https://giphy.com/gifs/fuck-you-middle-finger-ryan-stiles-x1kS7NRIcIigU")
    raise RuntimeError("Le joueur est trop con, je peux rien faire")


# Récupère un input par réaction
async def wait_for_react_clic(bot, message, emojis={}, *, process_text=False,
                              text_filter=lambda s: True, post_converter=None, trigger_all_reacts=False, trigger_on_commands=False):
    """Ajoute des reacts à un message et attend que quelqu'un appuie sur une

    Args:
        bot (:class:`.LGBot`): bot connecté au serveur
        message (:class:`discord.Message`): message où ajouter les réactions
        emojis (:class:`list` | :class:`dict`): reacts à ajouter, éventuellement associés à une valeur qui sera retournée au clic sur l'emoji
        process_text (:class:`bool`): si ``True``, détecte aussi la réponse par message et retourne ledit message (défaut : ``False``)
        text_filter (:class:`function`\(:class:`str` -> :class:`bool`\)): si ``process_text``, ne réagit qu'aux messages pour lesquels ``text_filter(message)`` renvoie ``True`` (défaut : tous)
        post_converter (:class:`function`\(:class:`str` -> :class:`object`\)): si ``process_text`` et que l'argument est défini, le message détecté est passé dans cette fonction avant d'être renvoyé
        trigger_all_reacts (:class:`bool`): si ``True``, détecte l'ajout des toutes les réactions (et pas seulement celles dans ``emojis``) et renvoie l'emoji directement si il n'est pas dans ``emojis`` (défaut : ``False``)
        trigger_on_commands (:class:`bool`): passé directement à :func:`.wait_for_message`.

    Returns:
        :class:`str`, représentant

            - le nom de l'emoji si ``emojis`` est une liste et clic sur une des reacts, ou si ``trigger_all_reacts`` vaut ``True`` et ajout d'une autre react ;
            - le message reçu si ``process_text`` vaut ``True``, que ``post_converter`` n'est pas défini et réaction à un message
    OU
        :class:`object`, représentant

            - la valeur associée si ``emojis`` est un dictionnaire et clic sur une des reacts ;
            - la valeur retournée par ``post_converter`` si il est défini, que ``process_text`` vaut ``True`` et réaction à un message
    """

    if not isinstance(emojis, dict):        # Si emoji est une liste, on en fait un dictionnaire
        emojis = {emoji: emoji for emoji in emojis}

    try:    # Si une erreur dans ce bloc, on supprime les emojis du bot (sinon c'est moche)
        for emoji in emojis:                    # On ajoute les emojis
            await message.add_reaction(emoji)

        emojis_names = {emoji.name if hasattr(emoji, "name") else emoji: emoji for emoji in emojis}
        def react_check(react):                     # Check REACT : bon message, pas un autre emoji, et pas react du bot
            return (react.message_id == message.id
                    and react.user_id != bot.user.id
                    and (trigger_all_reacts or react.emoji.name in emojis_names))

        react_task = asyncio.create_task(bot.wait_for('raw_reaction_add', check=react_check), name="react")

        if process_text:
            def message_check(mess):        # Check MESSAGE : bon channel, pas du bot, et filtre
                return (mess.channel == message.channel
                        and mess.author != bot.user
                        and text_filter(mess.content))
        else:
            def message_check(mess):        # On process DANS TOUS LES CAS, mais juste pour détecter "stop" si process_text == False
                return False

        mess_task = asyncio.create_task(wait_for_message(bot, check=message_check, trigger_on_commands=True), name="mess")

        done, pending = await asyncio.wait([react_task, mess_task], return_when=asyncio.FIRST_COMPLETED)      # On lance
        # Le bot attend ici qu'une des deux tâches aboutissent
        done_task = next(iter(done))        # done = tâche réussie

        if done_task.get_name() == "react":
            emoji = done_task.result().emoji
            if trigger_all_reacts and emoji.name not in emojis_names:
                ret = emoji
            else:
                ret = emojis.get(emoji) or emojis.get(emojis_names.get(emoji.name))     # Si clic sur react, done.result = react

            for emoji in emojis:
                await message.remove_reaction(emoji, bot.user)      # On finit par supprimer les emojis mis par le bot

        else:   # Réponse par message / STOP
            mess = done_task.result().content                # Si envoi de message, done.result = message
            ret = post_converter(mess) if post_converter else mess
            await message.clear_reactions()

    except Exception as exc:
        await message.clear_reactions()
        raise exc from Exception

    return ret


# Surcouche de wait_for_react_clic pour demander une confirmation / question fermée simplement
async def yes_no(bot, message):
    """Ajoute les reacts ✅ et ❎ à message et renvoie ``True`` ou ``False`` en fonction de l'emoji cliqué OU de la réponse textuelle détectée

    Surcouche de :func:`wait_for_react_clic` pour demander une confirmation / question fermée simplement.

    Args:
        bot (:class:`.LGBot`): bot connecté au serveur
        message (:class:`discord.Message`): message où ajouter les réactions

    Réponses textuelles reconnues :
        - Pour ``True`` : ``["oui", "o", "yes", "y", "1", "true"]``
        - Pour ``False`` : ``["non", "n", "no", "n", "0", "false"]``

    ainsi que toutes leurs variations de casse.

    Returns:
        :class:`bool`
    """
    yes_words = ["oui", "o", "yes", "y", "1", "true"]
    yes_no_words = yes_words + ["non", "n", "no", "n", "0", "false"]
    return await wait_for_react_clic(
        bot, message, emojis={"✅": True, "❎": False}, process_text=True,
        text_filter=lambda s: s.lower() in yes_no_words, post_converter=lambda s: s.lower() in yes_words)


# Surcouche de wait_for_react_clic pour demander de choisir dans une liste simplement
async def choice(bot, message, N, start=1, *, additionnal={}):
    """Ajoute des reacts chiffres (1️⃣, 2️⃣, 3️⃣...) à message et renvoie le numéro cliqué OU détecté par réponse textuelle

    Surcouche de :func:`wait_for_react_clic` pour demander de choisir dans une liste simplement.

    Args:
        bot (:class:`.LGBot`): bot connecté au serveur
        message (:class:`discord.Message`): message où ajouter les réactions
        N (:class:`int`): chiffre jusqu'auquel aller, inclus (``<= 10``)
        start (:class:`int`): chiffre auquel commencer (``<= N``, défaut ``1``)
        additionnal (:class:`dict`): dictionnaire emoji à ajouter après les chiffres -> valeur renvoyée si cliqué

    Réponses textuelles reconnues : nombres seuls entre ``start`` et ``N``

    Returns:
        :class:`int`
    """
    emojis = {emoji_chiffre(i): i for i in range(start, N+1)}
    if additionnal:
        emojis.update(additionnal)
    return await wait_for_react_clic(
        bot, message, emojis=emojis, process_text=True,
        text_filter=lambda s: s.isdigit() and start <= int(s) <= N, post_converter=int)


async def sleep(chan, x):
    """Pause l'exécution d'une commande en affichant l'indicateur *typing* ("*LGBot est en train d'écrire...*") sur un salon

    Permat d'afficher plusieurs messages d'affillée en laissant le temps de lire, tout en indiquant que le bot n'a pas fini d'écrire

    Args:
        chan (:class:`discord.abc.Messageable`): salon / contexte /... sur lequel attendre
        x (:class:`float`): temps à attendre, en secondes
    """
    async with chan.typing():
        await asyncio.sleep(x)



### ---------------------------------------------------------------------------
### Utilitaires d'emojis
### ---------------------------------------------------------------------------

def emoji_camp(arg, camp):
    """Renvoie l'emoji associé à un camp donné.

    Args:
        arg (:class:`~discord.ext.commands.Context` | :class:`~discord.Guild` | :class:`~discord.Member` | :class:`~discord.abc.GuildChannel`): argument "connecté" au serveur, permettant de remonter aux emojis
        camp (str): parmis ``"village"``, ``"loups"``, ``"nécro"``, ``"solitaire"``, ``"autre"``

    Returns:
        :class:`discord.Emoji` or `""`
    """
    d = {"village": "village",
         "loups": "lune",
         "nécro": "necro",
         "solitaire": "pion",
         "autre": "pion"}
    if camp in d:
        return emoji(arg, d[camp])
    else:
        return ""


def montre(heure=None):
    """Renvoie l'emoji horloge (🕧, 🕓, 🕝...) le plus proche d'une heure donnée

    Args:
        heure (:class:`str`): heure à représenter au format ``"XXh"`` ou ``"XXhMM"`` (défaut : heure actuelle)

    Returns:
        :class:`str`
    """
    if heure and isinstance(heure, str):
        heure, minute = heure.split("h")
        heure = int(heure) % 12
        minute = int(minute) % 60 if minute else 0
    else:
        now = datetime.datetime.now()
        heure = now.hour % 12
        minute = now.minute

    if 15 < minute < 45:        # Demi heure
        L = ["🕧", "🕜", "🕝", "🕞", "🕟", "🕠", "🕡", "🕢", "🕣", "🕤", "🕥", "🕦"]
    else:                       # Heure pile
        L = ["🕛", "🕐", "🕑", "🕒", "🕓", "🕔", "🕕", "🕖", "🕗", "🕘", "🕙", "🕚"]
    return L[heure] if minute < 45 else L[(heure + 1) % 12]


def emoji_chiffre(chiffre, multi=False):
    """Renvoie l'emoji / les emojis 0️⃣, 1️⃣, 2️⃣... correspond à un chiffre/nombre

    Args:
        chiffre (:class:`int`): chiffre/nombre à représenter
        multi (:class:`bool`): si ``True``, ``chiffre`` peut être n'importe quel entier positif, dont les chiffres seront convertis séparément ; sinon (par défaut), ``chiffre`` doit être un entier entre ``0`` et ``10``.

    Returns:
        :class:`str`
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 10:
        return ["0️⃣", "1️⃣", "2️⃣", "3️⃣", "4️⃣", "5️⃣", "6️⃣", "7️⃣", "8️⃣", "9️⃣", "🔟"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([emoji_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.emoji_chiffre doit être un entier entre 0 et 10 OU un entier positif avec multi=True")


def super_chiffre(chiffre, multi=False):
    """Renvoie le(s) caractère(s) Unicode ⁰, ¹, ²... correspond à un chiffre/nombre

    Args:
        chiffre (:class:`int`): chiffre/nombre à représenter
        multi (:class:`bool`): si ``True``, ``chiffre`` peut être n'importe quel entier positif, dont les chiffres seront convertis séparément ; sinon (par défaut), ``chiffre`` doit être un entier entre ``0`` et ``9``.

    Returns:
        :class:`str`
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["⁰", "¹", "²", "³", "⁴", "⁵", "⁶", "⁷", "⁸", "⁹"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([super_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.super_chiffre doit être un entier entre 0 et 9 OU un entier positif avec multi=True")


def sub_chiffre(chiffre, multi=False):
    """Renvoie le(s) caractère(s) unicode ₀, ₁, ₂... correspond à un chiffre/nombre

    Args:
        chiffre (:class:`int`): chiffre/nombre à représenter
        multi (:class:`bool`): si ``True``, ``chiffre`` peut être n'importe quel entier positif, dont les chiffres seront convertis séparément ; sinon (par défaut), ``chiffre`` doit être un entier entre ``0`` et ``9``.

    Returns:
        :class:`str`
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["₀", "₁", "₂", "₃", "₄", "₅", "₆", "₇", "₈", "₉"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([sub_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.sub_chiffre doit être un entier entre 0 et 9 OU un entier positif avec multi=True")



### ---------------------------------------------------------------------------
### Utilitaires de date / temps, notemment liées aux horaires de jeu
### ---------------------------------------------------------------------------

def heure_to_time(heure):
    """Convertit l'écriture d'une heure en objet :class:`datetime.time`

    Args:
        heure (:class:`str`): heure au format ``HHh``, ``HHhMM`` ou ``HH:MM``

    Returns:
        :class:`datetime.time`
    """
    try:
        if "h" in heure:
            hh, mm = heure.split("h")
        else:
            hh, mm = heure.split(":")
        return datetime.time(int(hh), int(mm) if mm else 0)
    except ValueError as exc:
        raise ValueError(f"Valeur \"{heure}\" non convertible en temps") from exc


def time_to_heure(tps, sep="h", force_minutes=False):
    """Convertit un objet :class:`datetime.time` en heure (version maison de :meth:`datetime.time.strftime`)

    Args:
        tps (:class:`datetime.time`): temps à convertir
        sep (:class:`str`): séparateur heures / minutes à utiliser (défaut ``"h"``)
        force_minutes (:class:`bool`): si ``False`` (défaut), les minutes ne sont indiquées que si différentes de ``0``.

    Returns:
        :class:`str` (``""`` si ``tps`` est ``None``)
    """
    if tps:
        if force_minutes or tps.minute > 0:
            return f"{tps.hour}{sep}{tps.minute:02}"
        else:
            return f"{tps.hour}{sep}"
    else:
        return ""


# Renvoie le datetime correspondant au prochain moment ou tps arrive DANS LES HORAIRES DU JEU : du dimanche 19:00:00 au vendredi 18:59:59.
def next_occurence(tps):
    """Renvoie le timestamp correspondant à la prochaine occurence d'une heure donnée

    Renvoie le prochain timestamp arrivant DANS LES HORAIRES DU JEU : du dimanche 19:00:00 au vendredi 18:59:59.

    Args:
        tps (:class:`datetime.time`): heure dont on veut connaître la prochaine occurence

    Returns:
        :class:`datetime.datetime`
    """
    pause = datetime.time(hour=19)

    now = datetime.datetime.now()
    jour = now.date()
    if tps <= now.time():       # Si plus tôt dans la journée que l'heure actuelle
        jour += datetime.timedelta(days=1)       # on réfléchit comme si on était demain très tôt

    wd = jour.weekday()         # Jour de la semaine, Lu = 0 ... Di = 6

    if tps < pause:
        if wd <= 4:                 # Avant 19h du lundi au vendredi : OK
            pass
        else:                       # Avant 19h mais on est samedi/dimanche
            jour += datetime.timedelta(days=(7-wd))
    else:
        if wd <= 3 or wd == 6:      # Après 19h du dimanche au jeudi : OK
            pass
        else:                       # Après 19h et on est vendredi/samedi
            jour += datetime.timedelta(days=(6-wd))

    return datetime.datetime.combine(jour, tps)         # passage de date et time à datetime


def debut_pause():
    """Renvoie le timestamp correspondant au prochain vendredi 19h

    Returns:
        :class:`datetime.datetime`
    """
    pause_time = datetime.time(hour=19)
    pause_wday = 4          # Vendredi

    now = datetime.datetime.now()
    jour = now.date()
    if pause_time <= now.time():        # Si plus tôt dans la journée que l'heure actuelle
        jour += datetime.timedelta(days=1)       # on réfléchit comme si on était demain très tôt

    pause_jour = jour + datetime.timedelta(days=(pause_wday - jour.weekday()) % 7)      # Jour décalé du nombre de jours avant vendredi
    return datetime.datetime.combine(pause_jour, pause_time)         # passage de date et time à datetime


def fin_pause():
    """Renvoie le timestamp correspondant au prochain dimanche 19h

    Returns:
        :class:`datetime.datetime`
    """
    reprise_time = datetime.time(hour=19)
    reprise_wday = 6        # Dimanche

    now = datetime.datetime.now()
    jour = now.date()
    if reprise_time <= now.time():      # Si plus tôt dans la journée que l'heure actuelle
        jour += datetime.timedelta(days=1)      # on réfléchit comme si on était demain très tôt

    reprise_jour = jour + datetime.timedelta(days=(reprise_wday - jour.weekday()) % 7)      # Jour décalé du nombre de jours avant vendredi
    return datetime.datetime.combine(reprise_jour, reprise_time)        # passage de date et time à datetime



### ---------------------------------------------------------------------------
### Split et log
### ---------------------------------------------------------------------------

# Sépare <mess> en une liste de messages de moins de <N>=2000 mots (limitation Discord), en séparant aux <sep>=sauts de ligne si possible.
# Ajoute <rep> à la fin des messages tronqués de leur séparateur final.
def smooth_split(mess, N=1990, sep='\n', rep=''):
    """Sépare un message en une liste de messages moins longs qu'un nombre de caractères donné

    Très utile pour envoyer des messages de (potentiellement) plus de 2000 caractères (limitation Discord)

    Args:
        mess (:class:`str`): message à couper
        N (:class:`int`): taille maximale des messages formés (défaut ``1990``, pour avoir un peu de marge et permettre d'entourer de ``````` par exemple)
        sep (:class:`str`) : caractères où séparer préférentiellement le texte (défaut : sauts de ligne). Si ``mess`` contient une sous-chaîne plus longue que ``N`` ne contenant pas ``sep``, le message est tronqué à la limite.
        rep (:class:`str`) : chaîne ajoutée à la fin de chaque message formé (tronqué du séparateur final) (défaut : aucune)

    Returns:
        :class:`list`\[:class:`str`\]
    """
    mess = str(mess)
    LM = []             # Liste des messages
    psl = 0             # indice du Précédent Saut de Ligne
    L = len(mess)
    while psl + N < L:
        if mess.count(sep, psl, psl+N+len(sep)):       # +len(sep) parce que si sep est à la fin, on le dégage
            i = psl + N - mess[psl: psl+N+len(sep)][::-1].find(sep)      # un peu sombre mais vrai, tkt frère
            LM.append(mess[psl: i] + rep)
            psl = i + 1     # on élimine le \n
        else:
            LM.append(mess[psl: psl + N])
            psl += N

    if psl < L:
        LM.append(mess[psl:])   # ce qui reste
    return LM


async def send_blocs(messageable, mess, *, N=1990, sep='\n', rep=''):
    """Envoie un (potentiellement long) message en le coupant en blocs si nécaissaire

    Surcouche de :func:`.smooth_split` envoyant directement les messages formés

    Retourne la liste des messages envoyés

    Args:
        messageable (:class:`discord.abc.Messageable`): objet où envoyer le message (:class:`~discord.ext.commands.Context` ou :class:`~discord.TextChannel`)
        mess (:class:`str`): message à envoyer
        N, sep, rep: *identique à* :func:`.smooth_split`

    Returns:
        :class:`list`\[:class:`discord.Message`\]
    """
    messages = []
    for bloc in smooth_split(mess, N=N, sep=sep, rep=rep):
        messages.append(await messageable.send(bloc))

    return messages


async def send_code_blocs(messageable, mess, *, N=1990, sep='\n', rep='', prefixe="", langage=""):
    """Envoie un (potentiellement long) message sous forme de bloc(s) de code

    Retourne la liste des messages envoyés

    :Paramètres:
        messageable, mess, N, sep, rep: *identiques à :func:`.send_blocs`*
        prefixe (:class:`str`): texte à mettre hors des code blocs, au début du premier message
        language: *identique à* :func:`.code_bloc`

    Returns:
        :class:`list`\[:class:`discord.Message`\]
    """
    if prefixe:
        prefixe = prefixe.rstrip() + "\n"

    messages = []
    for i, bloc in enumerate(smooth_split(prefixe + mess, N=N, sep=sep, rep=rep)):
        if prefixe and i == 0:
            bloc = bloc[len(prefixe):]
            message = await messageable.send(prefixe + code_bloc(bloc, langage=langage))
        else:
            message = await messageable.send(code_bloc(bloc, langage=langage))
        messages.append(message)

    return messages


# Log dans #logs
async def log(arg, message, *, code=False, N=1990, sep='\n', rep='', prefixe="", langage=""):
    """Envoie un message dans le channel ``#logs``

    Retourne la liste des messages envoyés

    Args:
        arg (:class:`~discord.ext.commands.Context` | :class:`~discord.Guild` | :class:`~discord.Member` | :class:`~discord.abc.GuildChannel`): argument "connecté" au serveur, permettant de remonter aux channels
        message (:class:`str`): message à log
        code (:class:`bool`): si ``True``, log sous forme de bloc(s) de code (défaut ``False``)
        N, sep, rep: *identique à* :func:`.send_blocs`
        prefixe: *identique à* :func:`.send_code_blocs`, simplement ajouté avant ``message`` si ``code`` vaut ``False``
        language: *identique à* :func:`.send_code_blocs`, sans effet si `code` vaut ``False``

    Returns:
        :class:`list`\[:class:`discord.Message`\]
    """
    logchan = channel(arg, "logs")
    if code:
        return (await send_code_blocs(logchan, message, N=N, sep=sep, rep=rep, prefixe=prefixe, langage=langage))
    else:
        if prefixe:
            message = prefixe.rstrip() + "\n" + message
        return (await send_blocs(logchan, message, N=N, sep=sep, rep=rep))



### ---------------------------------------------------------------------------
### Autres fonctions diverses
### ---------------------------------------------------------------------------

async def create_context(bot, member, content):
    """Simule qu'un membre a envoyé une message dans son chan privé et "génère" le contexte associé

    Args:
        bot (:class:`.LGBot`): bot connecté au serveur
        member (:class:`discord.Member`): membre dont on veut simuler l'action. **Doit être inscrit en base** (pour avoir un chan privé)
        content (:class:`str`): message à "faire envoyer" au joueur, généralement une commande


    Utile notemment pour simuler des commandes à partir de clics sur des réactions.

    Returns:
        :class:`discord.ext.commands.Context`
    """
    chan = private_chan(member)
    message = (await chan.history(limit=1).flatten())[0]        # On a besoin de récupérer un message, ici le dernier de la conv privée
    message.author = member
    message.content = content
    return await bot.get_context(message)


def nom_role(role, prefixe=False):
    """Retourne le nom d'un rôle à partir de son *slug*

    Args:
        role (:class:`str`): :attr:`.bdd.Roles.slug` à chercher
        prefile (:class:`bool`): inclure le préfixe ou non

    Returns:
        :class:`str` (:attr:`.bdd.Roles.slug`) | ``None`` (si non trouvé)
    """
    if role := Roles.query.get(role):
        if prefixe:
            return f"{role.prefixe}{role.nom}"
        else:
            return role.nom
    else:
        return None


# Remove accents
def remove_accents(s):
    """Enlève les accents d'un chaîne, mais conserve les caractères spéciaux non linguistiques (emojis...)

    Version plus douce de ``unidecode.unidecode``.

    Args:
        s (:class:`str`): chaîne à désaccentuer

    Returns:
        :class:`str`
    """
    p = re.compile("([À-ʲΆ-ת])")      # Abracadabrax, c'est moche mais ça marche (source : tkt frère)
    return p.sub(lambda c: unidecode.unidecode(c.group()), s)


# Évaluation d'accolades
def eval_accols(rep, globals_=None, locals_=None, debug=False):
    """Replace chaque bloc entouré par des ``{}`` par leur évaluation Python.

    Args:
        globals_ (:class:`dict`): variables globales du contexte d'évaluation (passé à :func:`eval`)
        locals_ (:class:`dict`): variables locales du contexte d'évaluation (passé à :func:`eval`)
        debug (:class:`bool`): si ``True``, insère le message d'erreur (type et texte de l'exception dans le message) ensuite si une exception est levée durant l'évaluation (défaut ``False``)

    Penser à passer les :func:`globals` et :func:`locals` si besoin. Généralement, il faut passer :func:`locals` qui contient ``ctx``, etc... mais pas :func:`globals` si on veut bénéficier de tous les modules importés dans ``tools.py`` (tous les modules du projet ou presque).
    """
    if globals_ is None:
        globals_ = globals()
    if locals_ is None:
        locals_ = globals_

    if "{" in rep:              # Si contient des expressions
        evrep = ""                  # Réponse évaluée
        expr = ""                   # Expression à évaluer
        noc = 0                     # Nombre de { non appariés
        for c in rep:
            if c == "{":
                if noc:             # Expression en cours :
                    expr += c           # on garde le {
                noc += 1
            elif c == "}":
                noc -= 1
                if noc:             # idem
                    expr += c
                else:               # Fin d'une expression
                    try:                                            # On essaie d'évaluer la chaîne
                        evrep += str(eval(expr, globals_, locals_))       # eval("expr") = expr
                    except Exception as e:
                        evrep += "{" + expr + "}"                   # Si erreur, on laisse {expr} non évaluée
                        if debug:
                            evrep += code(f"->!!! {e} !!!")
                    expr = ""
            elif noc:               # Expression en cours
                expr += c
            else:                   # Pas d'expression en cours
                evrep += c
        if noc:     # Si expression jamais finie (nombre impair de {)
            evrep += "{" + expr
        return evrep
    else:
        return rep



### ---------------------------------------------------------------------------
### Utilitaires de formatage de texte
### ---------------------------------------------------------------------------

def bold(s):
    """Formate une chaîne comme texte en **gras** dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f"**{s}**"

def ital(s):
    """Formate une chaîne comme texte en *italique* dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f"*{s}*"

def soul(s):
    """Formate une chaîne comme texte souligné dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f"__{s}__"

def strike(s):
    """Formate une chaîne comme texte barré dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f"~~{s}~~"

def code(s):
    """Formate une chaîne comme ``code`` (inline) dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f"`{s}`"

def code_bloc(s, langage=""):
    """Formate une chaîne comme un bloc de code dans Discord

    Args:
        s (:class:`str`): chaîne à formater
        langage (:class:`str`): langage du code, permet la coloration syntaxique (sur ordinateur uniquement)

    Langages supportés (non exhaustif ?) : ``asciidoc``, ``autohotkey``, ``bash``, ``coffeescript``, ``cpp`` (C++), ``cs`` (C#), ``css``, ``diff``, ``fix``, ``glsl``, ``ini``, ``json``, ``md``, (markdown), ``ml``, ``prolog``, ``py``, ``tex``, ``xl``, ``xml``

    Returns:
        :class:`str`
    """
    return f"```{langage}\n{s}```"

def quote(s):
    """Formate une chaîne comme citation (inline) dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f"> {s}"

def quote_bloc(s):
    """Formate une chaîne comme bloc de citation (multiline) dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f">>> {s}"

def spoiler(s):
    """Formate une chaîne comme spoiler (cliquer pour afficher) dans Discord

    Args:
        s (:class:`str`): chaîne à formater

    Returns:
        :class:`str`
    """
    return f"||{s}||"



"""lg-rez / features / Tâches planifiées

Planification, liste, annulation, exécution de tâches planifiées

"""

import datetime
import asyncio
import time

from discord.ext import commands

from lgrez.blocs import env, webhook, bdd, tools
from lgrez.blocs.bdd import Taches, Actions, Joueurs


_last_time = None       # Temps (time.time) du derner envoi de webhook

def execute(tache, loop):
    """Exécute une tâche planifiée

    Args:
        tache (:class:`.bdd.Taches`): tâche planifiée arrivée à exécution
        loop (:class:`asyncio.AbstractEventLoop`): boucle d'évènements du bot

    Envoie un webhook (variable d'environnement ``LGREZ_WEBHOOK_URL``) avec la commande (:attr:`.bdd.Taches.commande`) et nettoie

    Limitation interne de 2 secondes minimum entre deux appels (reprogramme si appelé trop tôt), pour se conformer à la rate limit Discord (30 messages / minute) et ne pas engoncer la loop
    """
    global _last_time

    if _last_time and (time.time() - _last_time) < 2:      # Moins de deux secondes depuis le dernier envoi
        loop.call_later(2, execute, tache, loop)        # on interdit l'envoi du webhook
        return

    _last_time = time.time()

    LGREZ_WEBHOOK_URL = env.load("LGREZ_WEBHOOK_URL")
    if webhook.send(tache.commande, url=LGREZ_WEBHOOK_URL):     # envoi webhook OK
        bdd.session.delete(tache)
        bdd.session.commit()
    else:
        # On réessaie dans 2 secondes
        loop.call_later(2, execute, tache, loop)


def add_task(bot, timestamp, commande, action=None):
    """Crée une nouvelle tâche planifiée sur le bot + en base, renvoie l'objet Tâche

    Args:
        bot (:class:`.LGBot`): bot connecté
        timestamp (:class:`datetime.datetime`): timestamp de la tâche à ajouter
        commande (:class:`str`): commande à exécuter à ``timestamp``
        action (:class:`int`): si tâche planifiée liée à une commande, son ID (:attr:`.bdd.Actions.id`)

    Returns:
        :class:`.bdd.Taches`

    (fonction pour usage ici et dans d'autres features)
    """
    now = datetime.datetime.now()
    tache = Taches(timestamp=timestamp, commande=commande, action=action)

    bdd.session.add(tache)                                                           # Enregistre la tâche en BDD
    bdd.session.commit()

    TH = bot.loop.call_later((timestamp - now).total_seconds(), execute, tache, bot.loop)     # Programme la tâche (appellera execute(tache, loop) à timestamp)
    bot.tasks[tache.id] = TH        # TaskHandler, pour pouvoir cancel

    return tache


def cancel_task(bot, tache):
    """Supprime (annule) une tâche (fonction pour usage ici et dans d'autres features)

    Args:
        bot (:class:`.LGBot`): bot connecté
        tache (:class:`.bdd.Taches`): tâche à annuler
    """
    bot.tasks[tache.id].cancel()        # Annulation (objet TaskHandler)
    bdd.session.delete(tache)            # Suppression en base
    bdd.session.commit()
    del bot.tasks[tache.id]             # Suppression TaskHandler


class GestionTaches(commands.Cog):
    """GestionTaches - Commandes de planification, exécution, annulation de tâches"""

    @commands.command()
    @tools.mjs_only
    async def taches(self, ctx):
        """Liste les tâches en attente (COMMANDE MJ)

        Affiche les commandes en attente d'exécution (dans la table :class:`.bdd.Taches`) et le timestamp d'exécution associé.
        Lorsque la tâche est liée à une action, affiche le nom de l'action et du joueur concerné.
        """
        async with ctx.typing():
            taches = Taches.query.order_by(Taches.timestamp).all()
            LT = ""
            for tache in taches:
                LT += f"\n{str(tache.id).ljust(5)} {tache.timestamp.strftime('%d/%m/%Y %H:%M:%S')}    {tache.commande.ljust(25)} "
                if tache.action and (action := Actions.query.get(tache.action)):
                    joueur = Joueurs.query.get(action.player_id)
                    assert joueur, f"!taches : joueur d'ID {action.player_id} introuvable"
                    LT += f"{action.action.ljust(20)} {joueur.nom}"

            mess = ("Tâches en attente : \n\nID    Timestamp              Commande                  Action               Joueur"
                    f"\n{'-'*105}{LT}\n\n"
                    "Utilisez !cancel <ID> pour annuler une tâche.") if LT else "Aucune tâche en attente."

        await tools.send_code_blocs(ctx, mess)


    @commands.command(aliases=["doat"])
    @tools.mjs_only
    async def planif(self, ctx, quand, *, commande):
        """Planifie une tâche au moment voulu (COMMANDE MJ)

        Args:
            quand: format ``[<J>/<M>[/<AAAA>]-]<H>:<M>[:<S>]``, avec ``<J>`` (jours), ``<M>`` (mois), ``<AAAA>`` (année sur 4 chiffres), ``<H>`` (heures) et ``<M>`` (minutes) des entiers et ``<S>`` (secondes) un entier ou un flottant, optionnel (défaut : ``0``)
                La date est optionnelle (défaut : date du jour). Si elle est précisée, elle doit être **séparée de l'heure par un tiret** et l'année peut être omise (défaut : année actuelle) ;
            commande: commande à exécuter (commençant par un ``!``). La commande sera exécutée PAR UN WEBHOOK dans LE CHAN ``#logs`` : toutes les commandes qui sont liées au joueur ou réservées au chan privé sont à proscrire (ou doivent a minima être précédées de ``!doas cible``)

        Cette commande repose sur l'architecture en base de données, ce qui garantit l'exécution de la tâche même si le bot plante entre temps.

        Si le bot est down à l'heure d'exécution prévue, la commande sera exécutée dès le bot de retour en ligne.

        Si la date est dans le passé, la commande est exécutée immédiatement.

        Examples:
            - ``!planif 18:00 !close maire``
            - ``!planif 13/06-10:00 !open maire``
            - ``!planif 13/06/2020-10:00 !open maire``
            - ``!planif 23:25:12 !close maire``
        """
        now = datetime.datetime.now()

        if "/" in quand:            # Date précisée
            date, time = quand.split("-")
            J, MA = date.split("/", maxsplit=1)
            day = int(J)
            if "/" in MA:           # Année précisée
                M, A = MA.split("/")
                month = int(M)
                year = int(A)
            else:
                month = int(MA)
                year = now.year
            date = datetime.date(year=year, month=month, day=day)
        else:
            date = now.date()
            time = quand

        H, MS = time.split(":", maxsplit=1)
        hour = int(H)
        if ":" in MS:               # Secondes précisées
            M, S = MS.split(":")
            minute = int(M)
            second = int(S)
        else:
            minute = int(MS)
            second = 0
        time = datetime.time(hour=hour, minute=minute, second=second)

        ts = datetime.datetime.combine(date, time)

        action_id = None            # ID de l'action associée à la tâche (utile pour propagation à la suppression de l'action)
        try:
            quoi, id = commande.split(" ")
            if quoi in ["!open", "!close", "!remind"]:
                action_id = int(id)
        except ValueError:
            pass

        if ts < datetime.datetime.now():
            mess = await ctx.send("Date dans le passé ==> exécution immédiate ! On valide ?")
            if not await tools.yes_no(ctx.bot, mess):
                await ctx.send("Mission aborted.")
                return

        tache = add_task(ctx.bot, ts, commande, action=action_id)
        await ctx.send(f"{tools.code(commande)} planifiée pour le {tools.code(ts.strftime('%d/%m/%Y %H:%M:%S'))}.\n{tools.code(f'!cancel {tache.id}')} pour annuler.")



    @commands.command(aliases=["retard", "doin"])
    @tools.mjs_only
    async def delay(self, ctx, duree, *, commande):
        """Exécute une commande après XhYmZs (COMMANDE MJ)

        Args:
            quand: format ``[<X>h][<Y>m][<Z>s]``, avec ``<X>`` (heures) et ``<Y>`` (minutes) des entiers et ``<Z>`` (secondes) un entier ou un flottant. Chacune des trois composantes est optionnelle, mais au moins une d'entre elle doit être présente ;
            commande: commande à exécuter (commençant par un ``!``). La commande sera exécutée PAR UN WEBHOOK dans LE CHAN ``#logs`` : toutes les commandes qui sont liées au joueur ou réservées au chan privé sont à proscrire (ou doivent a minima être précédées d'un ``!doas <cible>``)

        Cette commande repose sur l'architecture en base de données, ce qui garantit l'exécution de la commande même si le bot plante entre temps.

        Si le bot est down à l'heure d'exécution prévue, la commande sera exécutée dès le bot de retour en ligne.

        Examples:
            - ``!delay 2h !close maire``
            - ``!delay 1h30m !doas @moi !vote Yacine Oussar``
        """
        secondes = 0
        try:
            if "h" in duree.lower():
                h, duree = duree.split("h")
                secondes += 3600*int(h)
            if "m" in duree.lower():
                m, duree = duree.split("m")
                secondes += 60*int(m)
            if "s" in duree.lower():
                s, duree = duree.split("s")
                secondes += float(s)
        except Exception as e:
            raise commands.BadArgument("<duree>") from e

        if duree or not secondes:
            raise commands.BadArgument("<duree>")

        ts = datetime.datetime.now() + datetime.timedelta(seconds=secondes)
        action_id = None            # ID de l'action associée à la tâche (utile pour propagation à la suppression de l'action)
        try:
            quoi, id = commande.split(" ")
            if quoi in ["!open", "!close", "!remind"]:
                action_id = int(id)
        except ValueError:
            pass

        add_task(ctx.bot, ts, commande, action=action_id)

        await ctx.send(f"Commande {tools.code(commande)} planifiée pour le {tools.code(ts.strftime('%d/%m/%Y %H:%M:%S'))}")


    @commands.command()
    @tools.mjs_only
    async def cancel(self, ctx, *ids):
        """Annule une ou plusieurs tâche(s) planifiée(s) (COMMANDE MJ)

        Args:
            *ids: IDs des tâches à annuler, séparées par des espaces.

        Utiliser ``!taches`` pour voir la liste des IDs.
        """
        taches = [tache for id in ids if id.isdigit() and (tache := Taches.query.get(int(id)))]
        if taches:
            message = await ctx.send("Annuler les tâches :\n" + "\n".join([f" - {tools.code(tache.timestamp.strftime('%d/%m/%Y %H:%M:%S'))} > {tools.code(tache.commande)}" for tache in taches]))
            if await tools.yes_no(ctx.bot, message):
                for tache in taches:
                    cancel_task(ctx.bot, tache)

                await ctx.send("Tâche(s) annulée(s).")
            else:
                await ctx.send("Mission aborted.")
        else:
            await ctx.send(f"Aucune tâche trouvée.")



"""lg-rez / blocs / Interfaçage Google Sheets

Connection, récupération de classeurs, modifications (implémentation de https://pypi.org/project/gspread)
"""

import json

from lgrez.blocs import env
import gspread
from oauth2client.service_account import ServiceAccountCredentials


def connect(key):
    """Charge les credentials GSheets (variable d'environment ``LGREZ_GCP_CREDENTIALS``) et renvoie le classeur demandé

    Args:
        key (str): ID du classeur à charger (25 caractères)

    Returns:
        :class:`gspread.models.Spreadsheet`
    """
    # use creds to create a client to interact with the Google Drive API
    LGREZ_GCP_CREDENTIALS = env.load("LGREZ_GCP_CREDENTIALS")

    scope = ['https://spreadsheets.google.com/feeds']
    creds = ServiceAccountCredentials.from_json_keyfile_dict(json.loads(LGREZ_GCP_CREDENTIALS), scope)
    client = gspread.authorize(creds)

    # Open the workbook
    workbook = client.open_by_key(key)

    return workbook


def update(sheet, modifs):
    """Met à jour une feuille GSheets avec les modifications demandées

    Args:
        sheet (:class:`gspread.models.Worksheet`): la feuille à modifier
        modifs (:class:`list`\[\(:class:`int`, :class:`int`, :class:`object`\)\]): liste de tuples ``(ligne, colonne, valeur)``

    Les IDs sont indexés à partir de ``0`` (cellule ``A1`` en ``(0, 0)``.

    Le type de la nouvelle valeur sera interpreté par ``gspread`` pour donner le type GSheets adéquat à la cellule (texte, numérique, temporel...)
    """
    lm = max([l for (l, c, v) in modifs])       # ligne max de la zone à modifier
    cm = max([c for (l, c, v) in modifs])       # colonne max de la zone à modifier

    # Récupère toutes les valeurs sous forme de cellules gspread
    cells = sheet.range(1, 1, lm+1, cm+1)   # gspread indexe à partir de 1 (comme les gsheets)

    cells_to_update = []
    for (l, c, v) in modifs:
        cell = [cell for cell in cells if cell.col == c+1 and cell.row == l+1][0]    # on récup l'objet Cell correspondant aux coords à modifier
        if isinstance(v, int) and v > 10**14:
            cell.value = str(v)
        elif v is None:
            cell.value = ""
        else:
            cell.value = v              # cells : ([<L1C1>, <L1C2>, ..., <L1Ccm>, <L2C1>, <L2C2>, ..., <LlmCcm>]
        cells_to_update.append(cell)

    sheet.update_cells(cells_to_update)



"""Loup-Garou de la Rez (lg-rez)

Discord bot for organizing boisterous Werewolf RP games ESPCI-style.

See github.com/loic-simon/lg-rez for informations.
"""

__title__ = "lg-rez"
__author__ = "Loïc Simon, Tom Lacoma"
__license__ = "MIT"
__copyright__ = "Copyright 2020 Loïc Simon & Tom Lacoma – Club BD-Jeux × GRIs – ESPCI Paris - PSL"
__version__ = "1.2.0"

from lgrez import bot
LGBot = bot.LGBot                        # Accès direct à lgrez.LGBot

__all__ = ["LGBot"]



"""lg-rez / features / Communication

Envoi de messages, d'embeds...

"""

import os
import datetime
import traceback
from collections import Counter

import discord
from discord.ext import commands
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from lgrez.blocs import tools, env, gsheets
from lgrez.blocs.bdd import session, Joueurs, Actions, CandidHaro
from lgrez.features import gestion_actions


class Communication(commands.Cog):
    """Communication - Envoi de messages, d'embeds..."""

    current_embed = None

    @commands.command()
    @tools.mjs_only
    async def embed(self, ctx, key=None, *, val=None):
        """Prépare un embed (message riche) et l'envoie (COMMANDE MJ)

        Args:
            key: sous-commande (voir ci-dessous). Si omis, prévisualise le brouillon d'embed actuellement en préparation ;
            val: valeur associée. Pour les sous-commandes de construction d'élement, supprime ledit élément si omis.

        - Sous-commandes générales :
            - ``!embed create <titre>`` :          Créer un nouveau brouillon d'embed (un seul brouillon en parallèle, pour tous les utilisateurs)
            - ``!embed delete`` :                  Supprimer le brouillon d'embed
            - ``!embed preview``  :                Voir l'embed sans les rappels de commande
            - ``!embed post [#channel]`` :         Envoyer l'embed sur ``#channel`` (chan courant si omis)

        - Sous-commandes de construction d'éléments :
            - Éléments généraux :
                - ``!embed title [titre]``
                - ``!embed description [texte]``
                - ``!embed url [url*]``
                - ``!embed color [#ffffff]``      (barre de gauche, code hexadécimal)

            - Auteur :
                - ``!embed author [nom]``
                - ``!embed author_url [url*]``
                - ``!embed author_icon [url**]``

            - Talon :
                - ``!embed footer [texte]``
                - ``!embed footer_icon [url**]``

            - Images :
                - ``!embed image [url**]``        (grande image)
                - ``!embed thumb [url**]``        (en haut à droite)

            - Champs : syntaxe spéciale
                - ``!embed field <i> <skey> [val]``
                    - ``i`` :            Numéro du champ (commençant à ``0``). Si premier champ non existant, le crée.
                    - ``skey`` :
                        - ``name`` :       Nom du champ
                        - ``value`` :      Valeur du champ
                        - ``delete`` :     Supprime le champ

                Les champs sont (pour l'instant) forcément de type inline (côte à côte).

        \* Les URL doivent commencer par http(s):// pour être reconnues comme telles.

        ** Ces URL doivent correspondre à une image.
        """

        if val is None:
            val = discord.Embed.Empty

        emb = self.current_embed                # Récupération de l'embed (stocké dans le cog)

        direct = [                              # Attributs modifiables directement : emb.attr = value
            "title",
            "description",
            "url",
        ]
        method = {                              # Attributs à modifier en appelant une méthode : emb.set_<attr>(value)
            "footer": ("footer", "text"),                           # method[key] = (<attr>, <value>)
            "footer_icon": ("footer", "icon_url"),
            "image": ("image", "url"),
            "thumb": ("thumbnail", "url"),
            "author_url": ("author", "url"),
            "author_icon": ("author", "icon_url"),
        }

        if not emb:                             # Pas d'embed en cours
            if key == "create" and val:
                emb = discord.Embed(title=val)
            else:
                await ctx.send(f"Pas d'embed en préparation. {tools.code('!embed create <titre>')} pour en créer un.")
                return

        elif key in direct:                     # Attributs modifiables directement
            setattr(emb, key, val)

        elif key in method:                     # Attributs à modifier en appelant une méthode
            prop, attr = method[key]
            getattr(emb, f"set_{prop}")(**{attr: val})          # emb.set_<prop>(<attr>=val)

        elif key == "author":                   # Cas particulier
            emb.set_author(name=val) if val else emb.remove_author()

        elif key == "color":                    # Autre cas particulier : conversion couleur en int
            try:
                emb.color = eval(val.replace("#", "0x")) if val else discord.Embed.Empty
            except Exception:
                await ctx.send("Couleur invalide")
                return

        elif key == "field":                    # Cas encore plus particulier
            i_max = len(emb.fields)                 # N fields ==> i_max = N+1
            try:
                i, skey, val = val.split(" ", maxsplit=2)
                i = int(i)
                if i < 0 or i > i_max:
                    await ctx.send("Numéro de field invalide")
                    return
                if skey not in ["name", "value", "delete"]:
                    raise ValueError()
            except Exception:
                await ctx.send("Syntaxe invalide")
                return

            if i == i_max:
                if skey == "name":
                    emb.add_field(name=val, value=f"!embed field {i} value <valeur>")
                elif skey == "value":
                    emb.add_field(name=f"!embed field {i} name <nom>", value=val)
                # emb.add_field(*, name, value, inline=True)

            else:
                if skey == "name":
                    emb.set_field_at(i, name=val, value=emb.fields[i].name)
                elif skey == "value":
                    emb.set_field_at(i, name=emb.fields[i].name, value=val)
                else:
                    emb.remove_field(i)
                # emb.set_field_at(i, *, name, value, inline=True)

        elif key == "delete":
            self.current_embed = None
            await ctx.send(f"Supprimé. {tools.code('!embed create <titre>')} pour en créer un.")
            return

        elif key == "create":
            await ctx.send(f"Déjà un embed en cours de création. Utiliser {tools.code('!embed delete')} pour le supprimer.")

        elif key == "preview":
            await ctx.send("Prévisuatisation :", embed=emb)
            await ctx.send(f"Utiliser {tools.code('!embed post #channel')} pour publier l'embed.")
            return

        elif key == "post":
            if not val:     # channel non précisé
                await ctx.send(embed=emb)
            elif chan := tools.channel(ctx, val, must_be_found=False):
                await chan.send(embed=emb)
                await ctx.send("Et pouf !")
            else:
                await ctx.send(f"Channel inconnu. Réessaye en le mentionnant ({tools.code('#channel')})")
            return

        elif key is not None:
            await ctx.send(f"Option {key} incorrecte : utiliser {tools.code('!help embed')} pour en savoir plus.")
            return


        h_emb = emb.copy()
        if not emb.title:
            h_emb.title = "!embed title <titre>"
        if not emb.description:
            h_emb.description = "!embed description <description>"
        if not emb.footer:
            h_emb.set_footer(text="!embed footer <footer>")
        if not emb.author:
            h_emb.set_author(name="!embed author <auteur>")

        i_max = len(emb.fields)                 # N fields ==> i_max = N+1
        h_emb.add_field(name=f"!embed field {i_max} name <nom>", value=f"!embed field {i_max} value <nom>")

        await ctx.send("Embed en préparation :", embed=h_emb)
        await ctx.send(f"Utiliser {tools.code('!embed preview')} pour prévisualiser l'embed.\n"
                       f"Autres options : {tools.code('!embed color <#xxxxxx> / url <url> / image <url> / thumb <url> / author_url <url> / footer_icon <url>')}")

        self.current_embed = emb



    @commands.command(aliases=["tell"])
    @tools.mjs_only
    async def send(self, ctx, cible, *, message):
        """Envoie un message à tous ou certains joueurs (COMMANDE MJ)

        Args:
            cible: destinataires
            message: message, éventuellement formaté

        ``cible`` peut être :
            - ``all`` :                 Tous les joueurs inscrits, vivants et morts
            - ``vivants`` :             Les joueurs en vie
            - ``morts`` :               Les joueurs morts
            - ``<crit>=<filtre>`` :     Les joueurs répondant au critère ``Joueurs.<crit> == <filtre>``. ``crit`` peut être ``"nom"``, ``"chambre"``, ``"statut"``, ``"role"``, ``"camp"``... L'ensemble doit être entouré de guillements si ``filtre`` contient un espace.
            - *le nom d'un joueur*      (raccourci pour ``nom=X``, doit être entouré de guillements si nom + prénom)

        ``message`` peut contenir un ou plusieurs bouts de code Python à évaluer, entourés d'accolades.

        L'évaluation est faite séparément pour chaque joueur, ce qui permet de personnaliser le message grâce aux variables particulières dépendant du joueur :
            - ``joueur`` :          objet BDD du joueur recevant le message  ==> ``joueur.nom``, ``joueur.role``...
            - ``member`` :          objet discord.Member associé             ==> ``member.mention``
            - ``chan`` :            objet :class:`discord.TextChannel` du chan privé du joueur

        Attention :
            - ``ctx`` :             objet :class:`discord.ext.commands.Context` de ``!send``  ==> ``ctx.author`` = lanceur de la commande !!!

        Les différentes tables de données sont accessibles sous leur nom (``Joueurs``, ``Roles``...)

        Il est impossible d'appeller des coroutines (await) dans le code à évaluer.

        Examples:
            - ``!send all Bonsoir à tous c'est Fanta``
            - ``!send vivants Attention {member.mention}, derrière toi c'est affreux !``
            - ``!send "role=Servante Dévouée" Ça va vous ? Vous êtes bien {joueur.role} ?``
        """
        if cible == "all":
            joueurs = Joueurs.query.all()
        elif cible == "vivants":
            joueurs = Joueurs.query.filter_by(statut="vivant").all()
        elif cible == "morts":
            joueurs = Joueurs.query.filter_by(statut="mort").all()
        elif "=" in cible:
            crit, filtre = cible.split("=", maxsplit=1)
            if hasattr(Joueurs, crit):
                joueurs = Joueurs.query.filter(getattr(Joueurs, crit) == filtre).all()
            else:
                await ctx.send(f"Critère \"{crit}\" incorrect. !help {ctx.invoked_with} pour plus d'infos.")
                return
        else:
            joueurs = [await tools.boucle_query_joueur(ctx, cible, "À qui ?")]

        if not joueurs:
            await ctx.send(f"Aucun joueur trouvé.")
            return

        await ctx.send(f"{len(joueurs)} trouvé(s), envoi...")
        for joueur in joueurs:
            member = ctx.guild.get_member(joueur.discord_id)
            chan = ctx.guild.get_channel(joueur.chan_id_)

            assert member, f"!send : Member associé à {joueur} introuvable"
            assert chan, f"!sed : Chan privé de {joueur} introuvable"

            evaluated_message = tools.eval_accols(message, locals_=locals())
            await chan.send(evaluated_message)

        await ctx.send(f"Fini.")



    @commands.command()
    @tools.mjs_only
    async def post(self, ctx, chan, *, message):
        """Envoie un message dans un salon (COMMANDE MJ)

        Args:
            chan: nom du salon ou sa mention
            message: message à envoyer (peut être aussi long que nécessaire, contenir des sauts de lignes...)
        """
        chan = tools.channel(ctx, chan)
        await chan.send(message)
        await ctx.send(f"Fait.")



    @commands.command()
    @tools.mjs_only
    async def plot(self, ctx, type):
        """Trace le résultat du vote et l'envoie sur #annonces (COMMANDE MJ)

        Args:
            type: peut être

                - ``cond``   pour le vote pour le condamné
                - ``maire``  pour l'élection à la Mairie

        Trace les votes sous forme d'histogramme à partir du Tableau de bord, en fait un embed en présisant les résultats détaillés et l'envoie sur le chan ``#annonces``.

        Si ``type == "cond"``, déclenche aussi les actions liées au mot des MJs.
        """
        class Cible():
            def __init__(self, nom, votes=0):
                self.nom = nom
                self.label = self.nom.replace(" ", "\n", 1)

                self.votes = votes

                self.votants = []

                self.joueur = Joueurs.query.filter_by(nom=nom).one()
                if not self.joueur:
                    raise ValueError(f"Joueur \"{nom}\" non trouvé en base")

                self.eligible = bool(CandidHaro.query.filter_by(type=haro_candidature, player_id=self.joueur.discord_id).all())

            def __repr__(self):
                return f"{self.nom} ({self.votes})"

            def __eq__(self, other):
                return isinstance(other, Cible) and self.nom == other.nom

            def set_votants(self, raw_votants):
                votants = [rv or "zzz" for rv in raw_votants]
                votants.sort()
                self.votants = ["Corbeau" if nom == "zzz" else nom for nom in votants]         # On trie par ordre alphabétique en mettant les corbeaux (= pas de votant) à la fin

            def couleur(self, choisi):
                if self == choisi:
                    return hex(couleur).replace("0x", "#")
                if self.eligible:
                    return "#64b9e9"
                else:
                    return "gray"


        try:
            if type == "cond":
                colonne_cible = "CondamnéRéel"
                colonne_votant = "VotantCond"
                haro_candidature = "haro"
                typo = "bûcher du jour"
                mort_election = "Mort"
                pour_contre = "contre"
                emoji = "bucher"
                couleur = 0x730000
                couleur_txt = "#730000"

            elif type == "maire":
                colonne_cible = "MaireRéel"
                colonne_votant = "VotantMaire"
                haro_candidature = "candidature"
                typo = "nouveau maire"
                mort_election = "Élection"
                pour_contre = "pour"
                emoji = "maire"
                couleur = 0xd4af37
                couleur_txt = "#d4af37"

            else:
                await ctx.send("Merci de spécifier les résultats à tracer parmi 'maire' et 'cond'")
                return


            await ctx.send("Récupération des votes...")
            async with ctx.typing():
            # Sorcellerie sur la feuille gsheets pour trouver la colonne "CondamnéRéel"

                workbook = gsheets.connect(env.load("LGREZ_TDB_SHEET_ID"))    # Tableau de bord
                sheet = workbook.worksheet("Journée en cours")
                values = sheet.get_all_values()         # Liste de liste des valeurs des cellules
                NL = len(values)

                head = values[2]            # Ligne d'en-têtes (noms des colonnes) = 3e ligne du TDB
                ind_col_cible = head.index(colonne_cible)
                ind_col_votants = head.index(colonne_votant)

                cibles_brutes = [val for i in range(3, NL) if (val := values[i][ind_col_cible])]
                nb_votes = len(cibles_brutes)

                cibles = [Cible(nom, votes) for (nom, votes) in Counter(cibles_brutes).most_common()]       # Liste des cibles (vérifie l'éligibilité...) triées du plus au moins votées
                for cible in cibles:        # Récupération votants
                    cible.set_votants([values[i][ind_col_votants] for i in range(3, NL) if values[i][ind_col_cible] == cible.nom])


            choisi = None
            eligibles = [c for c in cibles if c.eligible]

            if eligibles:
                maxvotes = eligibles[0].votes
                egalites = [c for c in eligibles if c.votes == maxvotes]

                if len(egalites) > 1:       # Égalité
                    mess = await ctx.send("Égalité entre\n" + "\n".join(f"{tools.emoji_chiffre(i+1)} {c.nom}" for i, c in enumerate(egalites)) + "\nQui meurt / est élu ? (regarder vote du maire, 0️⃣ pour personne / si le vainqueur est garde-loupé, inéligible ou autre)")
                    choice = await tools.choice(ctx.bot, mess, len(egalites), start=0)
                    if choice:      # pas 0
                        choisi = eligibles[choice-1]

                else:
                    mess = await ctx.send(f"Joueur éligible le plus voté : {tools.bold(eligibles[0].nom)}\nÇa meurt / est élu ? (pas garde-loupé, inéligible ou autre)")
                    if await tools.yes_no(ctx.bot, mess):
                        choisi = eligibles[0]


            # Paramètres plot
            discord_gray = '#2F3136'
            plt.figure(facecolor=discord_gray)
            plt.rcParams.update({'font.size': 16})
            ax = plt.axes(facecolor='#8F9194') #coloration de TOUT le graphe
            ax.tick_params(axis='both', colors='white')
            ax.spines['bottom'].set_color('white')
            ax.spines['left'].set_color(discord_gray)
            ax.spines['right'].set_color(discord_gray)
            ax.spines['top'].set_color(discord_gray)
            ax.set_facecolor(discord_gray)
            ax.set_axisbelow(True)
            ax.yaxis.set_major_locator(MaxNLocator(integer=True))

            # Plot
            ax.bar(x=range(len(cibles)),
                   height=[c.votes for c in cibles],
                   tick_label=[c.label for c in cibles],
                   color=[c.couleur(choisi) for c in cibles],
            )
            plt.grid(axis="y")

            if not os.path.isdir("figures"):
                os.mkdir("figures")

            image_path = f"figures/hist_{datetime.datetime.now().strftime('%Y-%m-%d')}_{type}.png"
            plt.savefig(image_path, bbox_inches="tight")


            ### --------------- Partie Discord ---------------

            # Détermination rôle et camp
            emoji_camp = None
            if choisi:
                if type == "cond":
                    role = tools.nom_role(choisi.joueur.role, prefixe=True)
                    mess = await ctx.send(f"Rôle à afficher pour {choisi.nom} = {role} ? (Pas double peau ou autre)")
                    if await tools.yes_no(ctx.bot, mess):
                        emoji_camp = tools.emoji_camp(ctx, choisi.joueur.camp)
                    else:
                        await ctx.send("Rôle à afficher :")
                        role = (await tools.wait_for_message_here(ctx)).content
                        mess = await ctx.send("Camp :")
                        emoji_camp = await tools.wait_for_react_clic(ctx.bot, mess, [tools.emoji_camp(ctx, camp) for camp in ["village", "loups", "nécro", "solitaire", "autre"]])

                    nometrole = f"{tools.bold(choisi.nom)}, {role}"
                else:
                    nometrole = f"{tools.bold(choisi.nom)}"

            # Création embed
            embed = discord.Embed(
                title=f"{mort_election} de {nometrole if choisi else 'personne, bande de tocards'}",
                description=f"{nb_votes} votes au total",
                color=couleur
            )
            embed.set_author(name=f"Résultats du vote pour le {typo}", icon_url=tools.emoji(ctx, emoji).url)

            if emoji_camp:
                embed.set_thumbnail(url=emoji_camp.url)

            rd = "\n".join(("A" if cible.votes == 1 else "Ont") + f" voté {pour_contre} {cible.nom} : " + ", ".join(cible.votants) for cible in cibles)
            embed.set_footer(text=rd)

            file = discord.File(image_path, filename="image.png")
            embed.set_image(url="attachment://image.png")

            # Envoi
            mess = await ctx.send("Ça part ?\n", file=file, embed=embed)
            if await tools.yes_no(ctx.bot, mess):
                # Envoi du graphe
                file = discord.File(image_path, filename="image.png")       # Un objet File ne peut servir qu'une fois, il faut le recréer
                # embed.set_image(url="attachment://image.png")

                await tools.channel(ctx, "annonces").send("@everyone Résultat du vote ! :fire:", file=file, embed=embed)
                await ctx.send(f"Et c'est parti dans {tools.channel(ctx, 'annonces').mention} !")

                if type == "cond":
                    # Actions au mot des MJs
                    for action in Actions.query.filter_by(trigger_debut="mot_mjs").all():
                        await gestion_actions.open_action(ctx, action)

                    await ctx.send("(actions liées au mot MJ activées)")

        except Exception:
            await tools.send_code_blocs(ctx, traceback.format_exc())

        else:
            await ctx.send("Mission aborted.")
            self.current_embed = embed



    @commands.command()
    @tools.mjs_only
    async def annoncemort(self, ctx, *, victime=None):
        """Annonce un mort hors-vote (COMMANDE MJ)

        Args:
            victime: mort à annoncer

        Envoie un embed dans ``#annonces``
        """
        joueur = await tools.boucle_query_joueur(ctx, victime, "Qui est la victime ?")

        role = tools.nom_role(joueur.role, prefixe=True)
        mess = await ctx.send(f"Rôle à afficher pour {joueur.nom} = {role} ? (Pas double peau ou autre)")
        if await tools.yes_no(ctx.bot, mess):
            emoji_camp = tools.emoji_camp(ctx, joueur.camp)
        else:
            await ctx.send("Rôle à afficher :")
            role = (await tools.wait_for_message_here(ctx)).content
            mess = await ctx.send("Camp :")
            emoji_camp = await tools.wait_for_react_clic(ctx.bot, mess, [tools.emoji_camp(ctx, camp) for camp in ["village", "loups", "nécro", "solitaire", "autre"]])

        if joueur.statut == "MV":
            mess = await ctx.send("Annoncer la mort-vivance ?")
            if await tools.yes_no(ctx.bot, mess):
                role += " Mort-Vivant"
                emoji_camp = tools.emoji_camp(ctx, "nécro")

        await ctx.send("Contexte ?")
        desc = (await tools.wait_for_message_here(ctx)).content

        # Création embed
        embed = discord.Embed(
            title=f"Mort de {tools.bold(joueur.nom)}, {role}",
            description=desc,
            color=0x730000
        )
        embed.set_author(name=f"Oh mon dieu, quelqu'un est mort !")
        if emoji_camp:
            embed.set_thumbnail(url=emoji_camp.url)

        mess = await ctx.send("Ça part ?\n" + tools.ital(f"(si Non, l'embed est personnalisable via {tools.code('!embed')} puis envoyable à la main)"), embed=embed)
        if await tools.yes_no(ctx.bot, mess):
            # Envoi du graphe
            await tools.channel(ctx, "annonces").send("@everyone Il s'est passé quelque chose ! :scream:", embed=embed)
            await ctx.send(f"Et c'est parti dans {tools.channel(ctx, 'annonces').mention} !")

        else:
            await ctx.send("Mission aborted.")
            self.current_embed = embed



"""lg-rez / Commandes et autres fonctionnalités

Chaque module de ``lgrez.features`` implémente une fonctionnalité spécifique des LGBots. La majorité implémentent un *Cog* (:class:`discord.ext.commands.Cog`) contenant une ou plusieurs commandes Discord, mais peuvent aussi définir des fonctions pour un usage public.

.. note ::

    Les Cogs et leurs commandes peuvent être vus en envoyant ``!help`` à un LGBot fonctionnel (en possédant les rôles `MJ` et `Joueur en vie`)

"""

__all__ = ["actions_publiques", "annexe", "gestion_actions", "IA", "informations", "inscription", "open_close", "remplissage_bdd", "sync", "taches", "voter_agir"]
